import pytest

from pytagged import nline, _utils
from conftest import path_to_src_file_singles


@pytest.mark.parametrize(
    "tags",
    [(",debug"),
     ("skip, ,debug"),
     ("slow, ,debug,,benchmark")]
)
def test_pytagged_get_newlines_raise_err(tags: str):
    tags = tags.split(',')
    path = path_to_src_file_singles("hello.py")

    # call get_newlines with proper file arg, but with illegal tags
    # should raise ValueError
    with open(path) as f:
        with pytest.raises(ValueError):
            nline.get_newlines(f, *tags)


def test_pytagged_get_newlines(src_to_target_params):
    src_file, target_file = src_to_target_params[:2]
    tags = src_to_target_params[2:]
    print(tags)
    with open(target_file) as f:
        expected_lines = f.readlines()

    with open(src_file) as f:
        src_lines = f.readlines()
        f.seek(0)
        actual_lines = nline.get_newlines(f, *tags)
    _utils.pretty_print_title("ACTUAL")
    _utils.print_raw_lines(actual_lines)
    print('\n')

    _utils.pretty_print_title("EXPECTED")
    _utils.print_raw_lines(expected_lines)

    assert src_lines != expected_lines
    assert actual_lines == expected_lines



from os import get_terminal_size
import time
from typing import Iterable, Callable, Any, Tuple


def print_raw_lines(lines: Iterable[str]):
    for i, ln in enumerate(lines):
        print(f"|{i:3}| {repr(ln)[1:-1]}")


def pretty_print_title(title: str, width: int = 0,
                       padded_char: str = '=', span: bool = False):
    if width:
        print(f"{title:{padded_char}^{width}}")
        return

    width = len(title) + 20

    if span:
        try:
            term_width = get_terminal_size().columns
            width = term_width
        except OSError:
            pass
    print(f"{title:{padded_char}^{width}}")


def time_fn(fn: Callable) -> Callable:
    """Decorator/wrapper for to measure the execution of a callable.
    A function using this decorator will return a tuple whose
    first item is the result of the function itself, and the
    second item is the measured execution time.

    Args:
        fn (Callable): a callable whose execution time will be measured

    Returns:
        Callable: a wrapper that returns the result of the wrapped
            function and the measured execution time
    """
    def wrapper(*args: Any, **kwargs: Any) -> Tuple[Any, float]:

        st = time.monotonic()
        res = fn(*args, **kwargs)
        elapsed = time.monotonic() - st

        return res, elapsed
    return wrapper


def time_fn_only(fn: Callable) -> Callable:
    """Decorator/wrapper for to measure the execution of a callable.
    Unlike time_fn, a function using this decorator will only
    return the its measured execution time.

    Args:
        fn (Callable): a callable whose execution time will be measured

    Returns:
        Callable: a wrapper that returns the execution time of
            the wrapped function
    """
    def wrapper(*args: Any, **kwargs: Any) -> float:
        st = time.monotonic()
        fn(*args, **kwargs)
        return time.monotonic() - st

    return wrapper



import re
from typing import IO, List

POUND = '#'
BLOCK_START = "# block:"
BLOCK_END = "# end"
TRIPLE_QUOTE = '"""'
TRIPLE_QUOTE_SINGLE = "'''"


def get_newlines(file: IO, tag: str, *tags: str) -> List[str]:

    if not tag or tag == ' ':
        raise ValueError("Tags cannot be an empty string or a whitepsace")

    for t in tags:
        if not t or t == ' ':
            raise ValueError("Tags cannot be an empty string or a whitepsace")

    tags_match_str = tag

    if len(tags) > 0:
        tags_match_str += '|' + '|'.join(tags)

    line_split_rgx = re.compile(r"^(?!\s*$)(\s*)(.+)")
    block_start_rgx = re.compile(rf"{BLOCK_START} ({tags_match_str})")
    triple_quote_rgx = re.compile(rf"{TRIPLE_QUOTE}|{TRIPLE_QUOTE_SINGLE}")
    inline_rgx = re.compile(
        rf"^(?!{POUND})(?!.*({TRIPLE_QUOTE}|{TRIPLE_QUOTE_SINGLE})).*{POUND} ({tags_match_str})$")  # noqa: E501
    # monotonic_time = time.monotonic
    cur_block_start_idx = -1
    cur_triple_quote_start_idx = -1
    indices = set()
    block_pairs = []
    triple_quote_pairs = []
    matches = []
    newlines = []
    cmt_lines = []

    # line_scan_st = monotonic_time()
    # inline_scan_dur = 0
    # split_line_dur = 0
    for i, ln in enumerate(file):
        # st = monotonic_time()
        newlines.append(ln)
        matched = line_split_rgx.search(ln)
        # split_line_dur += 1000 * (monotonic_time() - st)
        matches.append(matched)
        if not matched:
            continue
        non_whitespace = matched.group(2)
        if non_whitespace[0] == POUND:
            cmt_lines.append(i)

        # st = monotonic_time()
        if inline_rgx.search(non_whitespace):
            indices.add(i)
        # inline_scan_dur += 1000 * (monotonic_time() - st)

        if non_whitespace == BLOCK_END:
            block_pairs.append((cur_block_start_idx, i))
            cur_block_start_idx = -1
            continue

        if non_whitespace in (TRIPLE_QUOTE, TRIPLE_QUOTE_SINGLE):
            triple_quote_pairs.append((cur_triple_quote_start_idx, i))
            cur_triple_quote_start_idx = -1
            continue

        if block_start_rgx.match(non_whitespace):
            cur_block_start_idx = i
            continue

        if triple_quote_rgx.match(non_whitespace):
            cur_triple_quote_start_idx = i
            continue

    """ line_scan_dur = 1000 * (monotonic_time() - line_scan_st)
    print(f"Time taken for scanning lines & computing indices: \
        {line_scan_dur:.2f} ms")
    print(f"Time taken for rgx splitting: {split_line_dur:.2f} ms")
    print(f"Time taken for inline rgx match: {inline_scan_dur:.2f} ms") """

    # update block indices
    for idx in block_pairs:
        start, end = idx
        if start == -1:
            continue
        indices |= {j for j in range(start + 1, end) if j not in cmt_lines}

    # remove triple quote block indices
    for idx in triple_quote_pairs:
        start, end = idx
        if start == -1:
            continue
        indices -= {j for j in range(start, end + 1)}

    for i, matched in enumerate(matches):
        if not matched:
            continue
        if i not in indices:
            continue
        newlines[i] = matched.expand(r"\g<1># \g<2>\n")

    return newlines



import argparse
from enum import Enum
import statistics
import sys
import textwrap
from os import path
from shutil import get_terminal_size
from pathlib import Path
from tempfile import TemporaryFile
from typing import Sequence, IO, Tuple

from pytagged import nline
from pytagged._utils import print_raw_lines, pretty_print_title, time_fn_only

PY_EXT = '.py'


class Mode(Enum):
    """Enum for PyTagged cli mode
    """
    DEFAULT = 0
    PRINTONLY = 1
    BENCHMARK = 2
    VERBOSE = 3


def printonly_single_file(io: IO, fname:str, *tags: str):
    newlines = nline.get_newlines(io, *tags)
    pretty_print_title(fname, span=True)
    print_raw_lines(newlines)
    pretty_print_title("EOF", span=True)


def printonly_files(io_seq: Sequence[Tuple[str, IO]], *tags: str):
    print_raw = print_raw_lines
    print_title = pretty_print_title
    get_newlines = nline.get_newlines

    def print_fn(io: IO):
        newlines = get_newlines(io, *tags)
        print_raw(newlines)

    for fname, io in io_seq:
        print_title(fname, span=True)
        print_fn(io)
        print_title("EOF", span=True)
        print('')


def benchmark_single_file(io: IO, fname: str, num_runs: int, *tags: str):
    file_content_og = io.read()
    io.seek(0)
    num_lines = len(list(io))
    io.seek(0)

    # wrap modify_single_file in time_fn_only
    timer = time_fn_only(modify_single_file)
    elapsed_times = []
    for _ in range(num_runs):
        # time modifying the file
        et = timer(io, *tags)
        elapsed_times.append(et)

        # restore the file content
        io.seek(0)
        io.truncate()
        io.write(file_content_og)
        io.seek(0)

    heading = "PERFORMANCE REPORT"
    footer = "END_REPORT"
    avg_run = statistics.mean(elapsed_times) * 1000
    median_run = statistics.median(elapsed_times) * 1000
    num_runs = len(elapsed_times)
    avg_per_line = sum(elapsed_times) / (num_runs * 1000) * 1000

    stats = {
        "Average time": avg_run,
        "Median time": median_run,
        "Average time per line": avg_per_line,
    }

    misc = {
        "File": fname,
        "Number of runs": num_runs,
        "Number of lines": num_lines,
    }

    pretty_print_title(heading, span=True)
    width = get_terminal_size().columns // 5

    for k, v in misc.items():
        print(f"{k:{width}} {v:>{width}}")
    for k, v in stats.items():
        print(f"{k:{width}} {v:>{width - 2}.4f}ms")

    pretty_print_title(footer, span=True)


def benchmark_files(io_seq: Sequence[Tuple[str, IO]], num_runs: int, *tags: str):
    og_contents = {}
    total_lines = 0
    for name, io in io_seq:
        og_contents[name] = io.read()
        io.seek(0)
        total_lines += len(list(io))
        io.seek(0)

    # wrap modify_single_file in time_fn_only
    timer = time_fn_only(modify_files)
    elapsed_times = []
    for _ in range(num_runs):
        # time modifying files
        et = timer(io_seq, *tags)
        elapsed_times.append(et)

        for name, io in io_seq:
            io.seek(0)
            io.truncate()
            io.write(og_contents[name])
            io.seek(0)

    heading = "PERFORMANCE REPORT"
    footer = "END_REPORT"
    avg_run = statistics.mean(elapsed_times) * 1000
    median_run = statistics.median(elapsed_times) * 1000
    num_runs = len(elapsed_times)
    avg_per_line = sum(elapsed_times) / (num_runs * 1000) * 1000

    stats = {
        "Average time": avg_run,
        "Median time": median_run,
        "Average time per line": avg_per_line,
    }
    misc = {
        "Number of files": len(og_contents.keys()),
        "Number of runs": num_runs,
        "Number of lines": total_lines
    }

    pretty_print_title(heading, span=True)
    width = get_terminal_size().columns // 5

    for k, v in misc.items():
        print(f"{k:{width}} {v:>{width}}")
    for k, v in stats.items():
        print(f"{k:{width}} {v:>{width - 2}.4f}ms")

    pretty_print_title(footer, span=True)


def modify_single_file_verbose(io: IO, fname: str, *tags: str):
    newlines = nline.get_newlines(io, *tags)
    io.seek(0)
    io.truncate()
    io.writelines(newlines)
    pretty_print_title(fname, span=True)
    print_raw_lines(newlines)
    pretty_print_title("EOF", span=True)


def modify_files_verbose(io_seq: Sequence[Tuple[str, IO]], *tags: str):
    print_raw = print_raw_lines
    print_title = pretty_print_title
    get_newlines = nline.get_newlines
    raw_str = []
    fnames = []

    def mod_file(f: IO):
        newlines = get_newlines(f, *tags)
        f.seek(0)
        f.truncate()
        f.writelines(newlines)
        raw_str.append(newlines)

    for fname, f_io in io_seq:
        mod_file(f_io)
        fnames.append(fname)

    for i, lines in enumerate(raw_str):
        print_title(fnames[i], span=True)
        print_raw(lines)
        print_title("EOF", span=True)
        print('')


def modify_single_file(io: IO, *tags: str):
    newlines = nline.get_newlines(io, *tags)
    io.seek(0)
    io.truncate()
    io.writelines(newlines)


def modify_files(io_seq: Sequence[Tuple[str, IO]], *tags: str):
    get_newlines = nline.get_newlines

    def mod_file(f: IO):
        newlines = get_newlines(f, *tags)
        f.seek(0)
        f.truncate()
        f.writelines(newlines)
    for i in io_seq:
        mod_file(i[1])


def copy_to_temp_file(pathobj: Path) -> IO:
    with pathobj.open() as f:
        src = f.read()

    # write newlines to temp file
    # pathstr, ext = path.splitext(str(pathobj))
    # ext = ".tmp"

    temp_file = TemporaryFile(mode="w+")
    temp_file.write(src)
    temp_file.seek(0)
    return temp_file


class NonPythonFileError(Exception):
    """Raise this if not '.py' extension
    """
    pass


def main():
    arg_parser = argparse.ArgumentParser(
        description="Comment out tagged code in your python code",
        add_help=False,
        formatter_class=argparse.RawTextHelpFormatter)

    # required args
    arg_parser.add_argument("path",
                            type=str,
                            nargs='?',
                            default='.',
                            help=textwrap.dedent("""\
                            path to python file(s),
                            if this is a directory, the program
                            will work on all files with the .py
                            extention within that directory.
                            Defaults to the current working dir.\n"""))

    arg_parser.add_argument("-t", "--tags",
                            type=str,
                            nargs='*',
                            required=True,
                            help=textwrap.dedent("""\
                            one or more 'tags', this tells the program
                            what to comment out.\n \n"""))

    # optional args
    arg_parser.add_argument("-h", "--help",
                            action="store_true",
                            help="Show help messages and exit.\n \n")

    modes = arg_parser.add_mutually_exclusive_group()
    modes.add_argument("-b", "--benchmark",
                       type=int,
                       help=textwrap.dedent("""\
                            Number of benchmark runs, if this is supplied
                            the program will run for N times, and print out
                            some performance statistics. Note that after PyTagged
                            is done, files will be restored to their original
                            content. Will also ignore the -v flag.\n \n"""))

    modes.add_argument("-p", "--printonly",
                       action="store_true",
                       help=textwrap.dedent("""\
                            Print only mode, if this flag is equivalent to the -v
                            flag but the program will not modify file(s).\n \n"""))

    modes.add_argument("-v", "--verbose",
                       action="store_true",
                       help=textwrap.dedent("""\
                            Verbose mode, if this flag is used,
                            the program will print out, line by line,
                            the raw string of the modified file."""))

    args = arg_parser.parse_args()

    if args.help:
        arg_parser.print_help()
        sys.exit(0)

    mode_int = 0
    if args.printonly:
        mode_int = 1
    elif args.benchmark:
        mode_int = 2
    elif args.verbose:
        mode_int = 3

    run_mode = Mode(mode_int)

    path_str = args.path
    path_obj = Path(path_str)
    tags = args.tags

    # block: develop
    # print("PyTagged running in dev mode")
    # print(f"cli mode: {run_mode.name.lower()}, using tags: {', '.join(tags)}")
    # print('')
    # end

    try:
        if path_obj.is_file():
            _, ext = path.splitext(path_str)
            if ext and ext != PY_EXT:
                raise NonPythonFileError(f"{path_str} is not a python file")

            if run_mode is Mode.DEFAULT:
                file_io = path_obj.open('r+')
                modify_single_file(file_io, *tags)

            elif run_mode is Mode.PRINTONLY:
                file_io = path_obj.open()
                printonly_single_file(file_io, path_str, *tags)

            elif run_mode is Mode.BENCHMARK:
                benchmark_runs = args.benchmark
                file_io = copy_to_temp_file(path_obj)
                benchmark_single_file(file_io, path_str, benchmark_runs, *tags)

            elif run_mode is Mode.VERBOSE:
                file_io = path_obj.open('r+')
                modify_single_file_verbose(file_io, path_str, *tags)

            file_io.close()
            sys.exit(0)

        if path_obj.is_dir():
            py_paths = list(path_obj.glob("**/*.py"))
            if not py_paths:
                print(f"No .py files found in {path_str}")
                sys.exit(0)

            if run_mode is Mode.DEFAULT:
                io_seq = [(str(p), p.open("r+")) for p in py_paths]
                modify_files(io_seq, *tags)

            elif run_mode is Mode.PRINTONLY:
                io_seq = [(str(p), p.open()) for p in py_paths]
                printonly_files(io_seq, *tags)

            elif run_mode is Mode.BENCHMARK:
                # copy the src files to temp files
                run_num = args.benchmark
                io_seq = [(str(p), copy_to_temp_file(p)) for p in py_paths]
                benchmark_files(io_seq, run_num, *tags)

            elif run_mode is Mode.VERBOSE:
                io_seq = [(str(p), p.open("r+")) for p in py_paths]
                modify_files_verbose(io_seq, *tags)

            for i in io_seq:
                i[1].close()
            sys.exit(0)

        if not path_obj.exists():
            raise FileNotFoundError(f"Error: {path_str} not found")

    except (OSError, IOError, PermissionError,
            NonPythonFileError, FileNotFoundError) as e:
        print(e)
        sys.exit(-1)



from pathlib import Path
import subprocess


from pytagged._utils import print_raw_lines, pretty_print_title


def test_cli_singles(cleanup_test_path_singles, src_to_target_params):
    src_path = src_to_target_params[0]
    target_path = src_to_target_params[1]
    tags = src_to_target_params[2:]

    cmd = ["pytag", src_path, "-t", *tags]
    print(cmd)
    subprocess.run(cmd, check=True)

    with open(src_path) as f:
        src_lines = f.readlines()

    with open(target_path) as f:
        target_lines = f.readlines()

    pretty_print_title(src_path)
    print_raw_lines(src_lines)
    print('')
    pretty_print_title(target_path)
    print_raw_lines(target_lines)
    print('')

    assert src_lines == target_lines


def test_cli_multiples(cleanup_test_path_multiples, src_to_target_params_multiples):

    src_path, target_path = src_to_target_params_multiples[:2]
    tags = src_to_target_params_multiples[2:]
    cmd = ["pytag", src_path, "-t", *tags]

    print(cmd)
    subprocess.run(cmd, check=True)

    expected_contents = {}
    for f in Path(target_path).glob(r"**/*.py"):
        fname = f.parts[-1]
        with f.open() as fin:
            expected_contents[fname] = fin.readlines()

    actual_contents = {}
    for f in Path(src_path).glob(r"**/*.py"):
        fname = f.parts[-1]
        with f.open() as fin:
            actual_contents[fname] = fin.readlines()

    for name in actual_contents:
        actual = actual_contents[name]
        expected = expected_contents[f"expected_{name}"]
        assert actual == expected
